@@ -46,6 +46,24 @@ int uriCompareRangeA(const UriTextRangeA * a, const UriTextRangeA * b);
 #define URI_TEST_IP_SIX_FAIL(x) ASSERT_TRUE(URI_FALSE == uri_TESTING_ONLY_ParseIpSixA(x "]"))
 #define URI_TEST_IP_SIX_PASS(x) ASSERT_TRUE(URI_TRUE == uri_TESTING_ONLY_ParseIpSixA(x "]"))
 
+#define URI_EXPECT_BETWEEN(candidate, first, afterLast)  \
+	EXPECT_TRUE((candidate >= first) && (candidate <= afterLast))
+
+#define URI_EXPECT_OUTSIDE(candidate, first, afterLast)  \
+	EXPECT_TRUE((candidate < first) || (candidate > afterLast))
+
+#define URI_EXPECT_RANGE_BETWEEN(range, uriFirst, uriAfterLast)  \
+	URI_EXPECT_BETWEEN(range.first, uriFirst, uriAfterLast);  \
+	URI_EXPECT_BETWEEN(range.afterLast, uriFirst, uriAfterLast)
+
+#define URI_EXPECT_RANGE_OUTSIDE(range, uriFirst, uriAfterLast)  \
+	URI_EXPECT_OUTSIDE(range.first, uriFirst, uriAfterLast);  \
+	URI_EXPECT_OUTSIDE(range.afterLast, uriFirst, uriAfterLast)
+
+#define URI_EXPECT_RANGE_EMPTY(range)  \
+	EXPECT_TRUE((range.first != NULL)  \
+			&& (range.afterLast != NULL)  \
+			&& (range.first == range.afterLast))
 
 namespace {
 	bool testDistinctionHelper(const char * uriText, bool expectedHostSet,
@@ -249,7 +267,7 @@ TEST(UriSuite, TestIpSixOverread) {
 
 		// NOTE: This string is designed to not have a terminator
 		char uriText[2 + 3 + 2 + 1 + 1];
-		strncpy(uriText, "//[::44.1", sizeof(uriText));
+		memcpy(uriText, "//[::44.1", sizeof(uriText));
 
 		EXPECT_EQ(uriParseSingleUriExA(&uri, uriText,
 				uriText + sizeof(uriText), &errorPos), URI_ERROR_SYNTAX);
@@ -293,7 +311,7 @@ TEST(UriSuite, TestUri) {
 		ASSERT_TRUE(0 == uriParseUriA(&stateA, "//user:pass@localhost/one/two/three"));
 		uriFreeUriMembersA(&uriA);
 
-		// ANSI and Unicode
+		// Both narrow and wide string version
 		ASSERT_TRUE(0 == uriParseUriA(&stateA, "http://www.example.com/"));
 		uriFreeUriMembersA(&uriA);
 		ASSERT_TRUE(0 == uriParseUriW(&stateW, L"http://www.example.com/"));
@@ -1080,6 +1098,19 @@ TEST(UriSuite, TestAddBase) {
 		// Bug related to absolutePath flag set despite presence of host
 		ASSERT_TRUE(testAddBaseHelper(L"http://a/b/c/d;p?q", L"/", L"http://a/"));
 		ASSERT_TRUE(testAddBaseHelper(L"http://a/b/c/d;p?q", L"/g/", L"http://a/g/"));
+
+		// GitHub issue #92
+		EXPECT_TRUE(testAddBaseHelper(L"http://a/b/c/../d;p?q", L"../..", L"http://a/"));
+		EXPECT_TRUE(testAddBaseHelper(L"http://a/b/c/../d;p?q", L"../../", L"http://a/"));
+
+		EXPECT_TRUE(testAddBaseHelper(L"http://a/b/../c/d;p?q", L"../..", L"http://a/"));
+		EXPECT_TRUE(testAddBaseHelper(L"http://a/b/../c/d;p?q", L"../../", L"http://a/"));
+
+		EXPECT_TRUE(testAddBaseHelper(L"http://a/../b/c/d;p?q", L"../..", L"http://a/"));
+		EXPECT_TRUE(testAddBaseHelper(L"http://a/../b/c/d;p?q", L"../../", L"http://a/"));
+
+		EXPECT_TRUE(testAddBaseHelper(L"http://a/b/c/d;p?q", L"../../..", L"http://a/"));
+		EXPECT_TRUE(testAddBaseHelper(L"http://a/b/c/d;p?q", L"../../../", L"http://a/"));
 }
 
 namespace {
@@ -1459,6 +1490,48 @@ TEST(UriSuite, TestNormalizeSyntaxComponents) {
 				URI_NORMALIZE_FRAGMENT));
 }
 
+TEST(UriSuite, TestNormalizeSyntaxPath) {
+	// These are from GitHub issue #92
+	EXPECT_TRUE(testNormalizeSyntaxHelper(
+			L"http://a/b/c/../../..",
+			L"http://a/",
+			URI_NORMALIZE_PATH));
+	EXPECT_TRUE(testNormalizeSyntaxHelper(
+			L"http://a/b/../c/../..",
+			L"http://a/",
+			URI_NORMALIZE_PATH));
+	EXPECT_TRUE(testNormalizeSyntaxHelper(
+			L"http://a/b/c/../../..",
+			L"http://a/",
+			URI_NORMALIZE_PATH));
+
+	// .. and these are related
+	EXPECT_TRUE(testNormalizeSyntaxHelper(
+			L"http://a/..",
+			L"http://a/",
+			URI_NORMALIZE_PATH));
+	EXPECT_TRUE(testNormalizeSyntaxHelper(
+			L"/..",
+			L"/",
+			URI_NORMALIZE_PATH));
+	EXPECT_TRUE(testNormalizeSyntaxHelper(
+			L"http://a/..///",
+			L"http://a///",
+			URI_NORMALIZE_PATH));
+	EXPECT_TRUE(testNormalizeSyntaxHelper(
+			L"http://a/..///..",
+			L"http://a//",
+			URI_NORMALIZE_PATH));
+	EXPECT_TRUE(testNormalizeSyntaxHelper(
+			L"a/b/c/../../..",
+			L"",
+			URI_NORMALIZE_PATH));
+	EXPECT_TRUE(testNormalizeSyntaxHelper(
+			L"a/b/../../c/..",
+			L"",
+			URI_NORMALIZE_PATH));
+}
+
 TEST(UriSuite, TestNormalizeCrashBug20080224) {
 		UriParserStateW stateW;
 		int res;
@@ -2143,6 +2216,139 @@ TEST(FreeUriMembersSuite, MultiFreeWorksFine) {
 	uriFreeUriMembersA(&uri);  // second time
 }
 
+namespace {
+	void testFreeUriMembersFreesHostText(const char *const uriFirst) {  // issue #121
+		const char *const uriAfterLast = uriFirst + strlen(uriFirst);
+		UriUriA uri;
+
+		EXPECT_EQ(uriParseSingleUriA(&uri, uriFirst, NULL), URI_SUCCESS);
+		EXPECT_EQ(uriMakeOwnerA(&uri), URI_SUCCESS);
+
+		EXPECT_EQ(uri.owner, URI_TRUE);
+		EXPECT_TRUE(uri.hostText.first);
+		EXPECT_TRUE(uri.hostText.afterLast);
+		EXPECT_NE(uri.hostText.first, uri.hostText.afterLast);
+		URI_EXPECT_RANGE_OUTSIDE(uri.hostText, uriFirst, uriAfterLast);
+
+		uriFreeUriMembersA(&uri);
+
+		EXPECT_FALSE(uri.hostText.first);
+		EXPECT_FALSE(uri.hostText.afterLast);
+
+		uriFreeUriMembersA(&uri);  // second time
+	}
+}  // namespace
+
+TEST(FreeUriMembersSuite, FreeUriMembersFreesHostTextIp4) {  // issue #121
+	testFreeUriMembersFreesHostText("//192.0.2.0");  // RFC 5737
+}
+
+TEST(FreeUriMembersSuite, FreeUriMembersFreesHostTextIp6) {  // issue #121
+	testFreeUriMembersFreesHostText("//[2001:db8::]");  // RFC 3849
+}
+
+TEST(FreeUriMembersSuite, FreeUriMembersFreesHostTextRegname) {  // issue #121
+	testFreeUriMembersFreesHostText("//host123.test");  // RFC 6761
+}
+
+TEST(FreeUriMembersSuite, FreeUriMembersFreesHostTextFuture) {  // issue #121
+	testFreeUriMembersFreesHostText("//[v7.X]");  // arbitrary IPvFuture
+}
+
+TEST(MakeOwnerSuite, MakeOwner) {
+	const char * const uriString = "scheme://user:pass@[v7.X]:55555/path/../path/?query#fragment";
+	UriUriA uri;
+	char * uriFirst = strdup(uriString);
+	const size_t uriLen = strlen(uriFirst);
+	char * uriAfterLast = uriFirst + uriLen;
+
+	EXPECT_EQ(uriParseSingleUriExA(&uri, uriFirst, uriAfterLast, NULL), URI_SUCCESS);
+
+	// After plain parse, all strings should point inside the original URI string
+	EXPECT_EQ(uri.owner, URI_FALSE);
+	URI_EXPECT_RANGE_BETWEEN(uri.scheme, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_BETWEEN(uri.userInfo, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_BETWEEN(uri.hostText, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_BETWEEN(uri.hostData.ipFuture, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_BETWEEN(uri.portText, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_BETWEEN(uri.pathHead->text, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_BETWEEN(uri.pathHead->next->text, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_BETWEEN(uri.pathHead->next->next->text, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_EMPTY(uri.pathHead->next->next->next->text);
+	EXPECT_TRUE(uri.pathHead->next->next->next->next == NULL);
+	URI_EXPECT_RANGE_BETWEEN(uri.query, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_BETWEEN(uri.fragment, uriFirst, uriAfterLast);
+
+	EXPECT_EQ(uriMakeOwnerA(&uri), URI_SUCCESS);
+
+	// After making owner, *none* of the strings should point inside the original URI string
+	EXPECT_EQ(uri.owner, URI_TRUE);
+	URI_EXPECT_RANGE_OUTSIDE(uri.scheme, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_OUTSIDE(uri.userInfo, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_OUTSIDE(uri.hostText, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_OUTSIDE(uri.hostData.ipFuture, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_OUTSIDE(uri.portText, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_OUTSIDE(uri.pathHead->text, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_OUTSIDE(uri.pathHead->next->text, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_OUTSIDE(uri.pathHead->next->next->text, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_EMPTY(uri.pathHead->next->next->next->text);
+	EXPECT_TRUE(uri.pathHead->next->next->next->next == NULL);
+	URI_EXPECT_RANGE_OUTSIDE(uri.query, uriFirst, uriAfterLast);
+	URI_EXPECT_RANGE_OUTSIDE(uri.fragment, uriFirst, uriAfterLast);
+
+	// Free originally used memory so we'd get violations on access with ASan
+	uriAfterLast = NULL;
+	free(uriFirst);
+	uriFirst = NULL;
+
+	// Can we recompose the URI without accessing any old freed memory?
+	int charsRequired;
+	EXPECT_EQ(uriToStringCharsRequiredA(&uri, &charsRequired), URI_SUCCESS);
+	EXPECT_TRUE((charsRequired >= 0) && (charsRequired >= static_cast<int>(uriLen)));
+	char * const uriRemake = new char[charsRequired + 1];
+	EXPECT_TRUE(uriRemake != NULL);
+	EXPECT_EQ(uriToStringA(uriRemake, &uri, charsRequired + 1, NULL), URI_SUCCESS);
+	EXPECT_TRUE(! strcmp(uriString, uriRemake));
+	delete [] uriRemake;
+
+	uriFreeUriMembersA(&uri);
+}
+
+namespace {
+	void testMakeOwnerCopiesHostText(const char *const uriFirst) {  // issue #121
+		const char *const uriAfterLast = uriFirst + strlen(uriFirst);
+		UriUriA uri;
+
+		EXPECT_EQ(uriParseSingleUriA(&uri, uriFirst, NULL), URI_SUCCESS);
+		EXPECT_EQ(uri.owner, URI_FALSE);
+		URI_EXPECT_RANGE_BETWEEN(uri.hostText, uriFirst, uriAfterLast);
+
+		EXPECT_EQ(uriMakeOwnerA(&uri), URI_SUCCESS);
+
+		EXPECT_EQ(uri.owner, URI_TRUE);
+		URI_EXPECT_RANGE_OUTSIDE(uri.hostText, uriFirst, uriAfterLast);
+
+		uriFreeUriMembersA(&uri);
+		uriFreeUriMembersA(&uri);  // tried freeing stack pointers before the fix
+	}
+} // namespace
+
+TEST(MakeOwnerSuite, MakeOwnerCopiesHostTextIp4) {  // issue #121
+	testMakeOwnerCopiesHostText("//192.0.2.0");  // RFC 5737
+}
+
+TEST(MakeOwnerSuite, MakeOwnerCopiesHostTextIp6) {  // issue #121
+	testMakeOwnerCopiesHostText("//[2001:db8::]");  // RFC 3849
+}
+
+TEST(MakeOwnerSuite, MakeOwnerCopiesHostTextRegname) {  // issue #121
+	testMakeOwnerCopiesHostText("//host123.test");  // RFC 6761
+}
+
+TEST(MakeOwnerSuite, MakeOwnerCopiesHostTextFuture) {  // issue #121
+	testMakeOwnerCopiesHostText("//[v7.X]");  // arbitrary IPvFuture
+}
+
 TEST(ParseIpFourAddressSuite, FourSaneOctets) {
 	unsigned char octetOutput[4];
 	const char * const ipAddressText = "111.22.3.40";
